
/************************************************************************
  Author:

  Jens Jrgen Nielsen
  e-mail:jjn2970@gmail.com
  http:\\www.corix.dk
  
  Terms of use:
  
  To use this code in a project or product please ask the author for 
  a license. For non-commercial use it is currently (2015) provided for 
  free (2015). This however may change, so please ask.

  To distribute the source code please do NOT change anything in this file. 
  Otherwise feel free to pass it on.
*************************************************************************/

#include <math.h>
#include <float.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <time.h>
#include <sys\timeb.h>

#include "gf256.h"
#include "rs255.h"


#define  maxIndex  10000L
#define  eventTableSize 40


static unsigned char debugFlag;



//void rs255syndroms(unsigned char b[], unsigned char s[], long n);

/* Calculate time in seconds from start. Resolution 1/1000 second.
 */
double timeUsed(struct _timeb *start)
{                               
    double seconds;
    struct _timeb   finish;

    _ftime_s(&finish);
    seconds = (double) finish.millitm;
    seconds = seconds - start->millitm;
    seconds = seconds / 1000;
    seconds = seconds + finish.time - start->time;
    
    return(seconds);
}

void polyPrint(unsigned char b[], long n)
{
    long            i,k;
	
	k = 0;
    for (i=0; i<n; i++) {
		if (k==0) {
			printf("\n %3ld ",i);
			k = 20;
		}
		printf("%3lx",b[i]);
		k = k-1;
	}
}

/*Compare the two arrays and return the number of non equal elements*/
long rs255compare(unsigned char x[], unsigned char y[], long n)
{
    long            i;
	unsigned char   a;

	a = 0;
    for (i=0; i<n; i++) {
		if (x[i] != y[i]) {
			a++;
//		    printf("\nerror in position %3ld %3ld %3ld ",i,x[i],y[i]);
		}
	}
    return(a);
}

/*Return an array of 255 elements, each element different from 0 with a probability of p.
  If different from 0 all elements are equally likely.
*/
long rs255random(unsigned char x[], float p)
{
    long            i,k;
	unsigned char   a;
    int             n;

	k = 0;
    a = 0;
    for (i=0; i<255; i++) {
        n = rand();
        if (n < (32768*p)) {
            n = rand();
            a = n / 128;
            if (a>0) k = k + 1;
//            printf("\nError in position %3ld %3ld",i,a);
        }
        else {
            a = 0;
        }
		x[i] = a;
	}
    return(k);
}

/*Update the array by moving the first non-zero element one up. If it is already at the
  second element move that one up and return the first one to the lowest possible index.
  If the second element is at the third then move that one up and so on ...
  Alternative description: we are running through the numbers from zero and up and select
  all with a certain number of bits.

  76543210
  00110010->00110100->00111000->01000011

  When we enter the routine with the last number (all non-zeroes at their highest position) 
  we return -1, otherwise the position of the first movable is returned.

  Algorithm: start from index 0 and go up. Each time a non-movable is found it is moved to 
  its base position. When a moveable is found it is moved and we exit.

  The number of different patterns created this way is C(n,k) where n is the number of
  positions and k is the number of non-zero elements, e.g. C(255,4)=172061505.
   
*/
long nextWord(unsigned char x[], int xSize)
{
    long            i;
    int             firstMovable, nonMovable;
	unsigned char   a;

	firstMovable = -1;
	nonMovable = 0;
	i = 0;

    do {
        if (x[i] != 0) {
            if ((x[i+1] == 0) && (i < (xSize-1))) {
				firstMovable = i;
				x[i+1] = x[i];
				x[i] = 0;
			}
			else {
				a = x[i];
				x[i] = 0;
				x[nonMovable] = a;
				nonMovable++;
			}
		}
		i++;
	} while ((firstMovable == -1) && (i < xSize));

    return(firstMovable);
}
/*Return an array of 255 elements, each element different from 0 with a probability of 
  (px+py). If different from 0 all elements are equally likely. A fraction of the non-
  zero elements (corresponding to the probability py) are termed erasures and a "1" in 
  the y-array points to the "erased" symbol.
*/
long rs255channel(unsigned char x[], unsigned char y[], float px, float py)
{
    long            i,k;
	unsigned char   a;
    int             n,m;

	k = 0;
    a = 0;
    for (i=0; i<255; i++) {
        x[i]=0;
        y[i]=0;
        n = rand();
        if (n < (32768*px)) {
            m = rand();
            x[i] = m / 128;
            if (x[i]>0) k = k + 1;
        }
        if (((32768*px) <= n) && (n < (32768*(px+py)))) {
            m = rand();
            x[i] = m / 128;
            if (x[i]>0) {
                y[i]=1;
            }
        }
	}
    return(k);
}

void randomErrorErasureTest()
{
    long            i,j,k;
    float           errorProbability, erasureProbability;
    long            paritySymbolCount, correctableErrorCount;
    unsigned char   x[256],y[256],z[256],e[256],u[256];
    long            errorsInserted, erasuresInserted;
    long            errorIndex, erasureIndex;
    long            errorsCorrected, errorsFound;
    long            errorEvent;
    long            correctableEventsCorrected;
    long            correctableEventsNotCorrected;
    long            notCorrectableEventsCorrected;
    long            notCorrectableEventsNotCorrected;
    long            eventCount[eventTableSize][eventTableSize][3];
    unsigned char   c[256];
	float           alpha;
	long            testCount;

    FILE *txtFile;

    paritySymbolCount = 32;
	testCount = 100000;

	alpha = 0.5; //alpha = 1 : only errors, alpha = 0: only erasures
    erasureProbability = (float) (1-alpha)*paritySymbolCount/255;
    errorProbability   = (float) 0.5*alpha*paritySymbolCount/255;

    correctableErrorCount = paritySymbolCount/2;
    initGF256tables();

    printf("\nPlease wait for the first 4000 simulations\n");        
    fopen_s(&txtFile, "random.txt", "w");
    fprintf(txtFile,"\nRS255 test:");

    for (i=0; i<eventTableSize; i++) {
        for (j=0; j<eventTableSize; j++) {
            for (k=0; k<3; k++) {
                eventCount[i][j][k] = 0;
			}
        }
    }

	correctableEventsCorrected = 0;
	correctableEventsNotCorrected = 0;
	notCorrectableEventsCorrected = 0;
	notCorrectableEventsNotCorrected = 0;

	debugFlag = 0;
    srand(1007);

    i=0;
	do {
        rs255random(x,1.0);
        rs255encode(x,paritySymbolCount);

        errorsInserted = rs255channel(e, u, errorProbability, erasureProbability);

        erasuresInserted=0;
        for (j=0; j<255; j++) {
            y[j] = e[j] ^ x[j];
            if (u[j] !=0) erasuresInserted++;
        }
       
        errorEvent = errorsInserted+erasuresInserted;
        if (errorEvent > 39) errorEvent = 39;
        
		errorIndex = errorsInserted;
		if (errorsInserted > (eventTableSize-1))	errorIndex = (eventTableSize-1);
		
		erasureIndex = erasuresInserted;
		if (erasuresInserted > (eventTableSize-1)) erasureIndex = (eventTableSize-1);

        errorsCorrected = rs255decode(y, u, z, paritySymbolCount);

        errorsFound = rs255compare(x,z,255);

        if (debugFlag==1) {
            polyPrint(e,255);
            debugFlag=0;
        }

        if (errorsFound!=0) {
            eventCount[errorIndex][erasureIndex][1]++;
//            printf("*");
            if ((2*errorsInserted+erasuresInserted) <= paritySymbolCount) {
				correctableEventsNotCorrected++;

                printf("\nErrors not corrected: %3ld %3ld %3ld %3ld", 
                       errorsInserted, erasuresInserted, errorsCorrected, errorsFound);

                printf("\nEnter to continue ");
                scanf_s( "%c", &c );
            }
			else notCorrectableEventsNotCorrected++;
        }
        else {
            eventCount[errorIndex][erasureIndex][0]++;
//            printf(".");
            if ((errorsInserted+erasuresInserted) != errorsCorrected) {
                printf("\nNumber of errors do not match:  %3ld %3ld %3ld", 
                       errorsInserted, erasuresInserted, errorsCorrected);
                printf("\nEnter to continue ");
                scanf_s( "%c", &c );
            }
            if ((2*errorsInserted+erasuresInserted) <= paritySymbolCount) correctableEventsCorrected++;
			else notCorrectableEventsCorrected++;

        }
        i++;
		
		if ((i & 0xfff) == 0) {
            printf("\nCorrectable events with all errors corrected     %6ld", correctableEventsCorrected);
            printf("\nNon correctable events with all errors corrected %6ld", notCorrectableEventsCorrected);
            printf("\nCorrectable events with error in decoding        %6ld", correctableEventsNotCorrected);
            printf("\nNon correctable events with error in decoding    %6ld", notCorrectableEventsNotCorrected);
			printf("\n");
        }
    
    
        if ((i & 0xfff) == 0) {
            fprintf(txtFile,"\nEvents with all errors corrected\n");
            fprintf(txtFile,"      ");
            for (j=0; j<eventTableSize; j++)
                fprintf(txtFile,"%6ld", j);
            fprintf(txtFile,"\n");
            for (j=0; j<eventTableSize; j++) {
                fprintf(txtFile,"%6ld", j);
                for (k=0; k<eventTableSize; k++) {
//                if ((j % 8)==0) printf("\n%3ld:",j);
                    fprintf(txtFile,"%6ld", eventCount[j][k][0]);
				}
                fprintf(txtFile,"\n");
			}
            fprintf(txtFile,"\nEvents with errors after decoding\n");
            fprintf(txtFile,"      ");
            for (j=0; j<eventTableSize; j++)
                fprintf(txtFile,"%6ld", j);
            fprintf(txtFile,"\n");
            for (j=0; j<eventTableSize; j++) {
                fprintf(txtFile,"%6ld", j);
                for (k=0; k<eventTableSize; k++) {
//                if ((j % 8)==0) printf("\n%3ld:",j);
                    fprintf(txtFile,"%6ld", eventCount[j][k][1]);
				}
                fprintf(txtFile,"\n");
			}
            fprintf(txtFile,"\n");
        }
    
    } while (i<testCount);

    fclose(txtFile);   

    printf("\nEnter to continue ");
    scanf_s( "%c", &c );

    printf("\n");
    printf("\nGF(256) test terminated.");
    printf("\n");
}    

void printEventTable(FILE *txtOutput, long eventArray[][eventTableSize][3]) 
{
    long    j,k;

    if (0 == 0) {
        fprintf(txtOutput,"\nEvents with all errors corrected\n");
        fprintf(txtOutput,"      ");
        for (j=0; j<eventTableSize; j++)
            fprintf(txtOutput,"%6ld", j);
        fprintf(txtOutput,"\n");
        for (j=0; j<eventTableSize; j++) {
            fprintf(txtOutput,"%6ld", j);
            for (k=0; k<eventTableSize; k++) {
//                if ((j % 8)==0) printf("\n%3ld:",j);
                fprintf(txtOutput,"%6ld", eventArray[j][k][0]);
			}
            fprintf(txtOutput,"\n");
		}
        fprintf(txtOutput,"\nEvents with errors after decoding\n");
        fprintf(txtOutput,"      ");
        for (j=0; j<eventTableSize; j++)
            fprintf(txtOutput,"%6ld", j);
        fprintf(txtOutput,"\n");
        for (j=0; j<eventTableSize; j++) {
            fprintf(txtOutput,"%6ld", j);
            for (k=0; k<eventTableSize; k++) {
//                if ((j % 8)==0) printf("\n%3ld:",j);
                fprintf(txtOutput,"%6ld", eventArray[j][k][1]);
			}
            fprintf(txtOutput,"\n");
		}
        fprintf(txtOutput,"\n");
    }
}

void systematicErrorErasureTest()
{
    long            i,j,k;
    long            maxErrorCount, maxErasureCount;
    long            paritySymbolCount, correctableErrorCount;
    unsigned char   x[256],y[256],z[256],e[256],u[256];
    long            errorsInserted, erasuresInserted;
    long            errorIndex, erasureIndex;
    long            errorsCorrected, errorsFound;
    long            correctableEventsCorrected;
    long            correctableEventsNotCorrected;
    long            notCorrectableEventsCorrected;
    long            notCorrectableEventsNotCorrected;
    long            notCorrectableEventsErrCorrected;
    long            eventCount[eventTableSize][eventTableSize][3];
    unsigned char   c[256];

    FILE *txtFile;

    paritySymbolCount =   8;
    maxErasureCount =     0;
    maxErrorCount   =     3;

    correctableErrorCount = paritySymbolCount/2;
    initGF256tables();

    printf("\nPlease wait for the first 4000 simulations\n");        
    fopen_s(&txtFile, "systematic.txt", "w");
    fprintf(txtFile,"\nSystematic error/erasure test:");

    for (i=0; i<eventTableSize; i++) {
        for (j=0; j<eventTableSize; j++) {
            for (k=0; k<3; k++) {
                eventCount[i][j][k] = 0;
			}
        }
    }

	correctableEventsCorrected = 0;
	correctableEventsNotCorrected = 0;
	notCorrectableEventsCorrected = 0;
	notCorrectableEventsNotCorrected = 0;
	notCorrectableEventsErrCorrected = 0;

	debugFlag = 0;
    srand(1007);

	errorsInserted = maxErrorCount;
    erasuresInserted=0;

    for (j=0; j<255; j++) {
        e[j] = 0;
		u[j] = 0;
        if (j<errorsInserted) e[j] = 1;
    }

/* Make errorpattern which is a codeword
    for (j=0; j<255; j++) {
        e[j] = 0;
		u[j] = 0;
    }
	e[0] = 1;
	rs255encode(e,paritySymbolCount);

    errorsInserted = 0;
	for (j=0; j<255; j++) {
        if (e[j] != 0) errorsInserted++;
    }

*/
//	polyPrint(e, 255);


    rs255random(x,1.0);
    rs255encode(x,paritySymbolCount);
	
	i=0;
	do {

        for (j=0; j<255; j++) {
            y[j] = e[j] ^ x[j];
        }
       
		errorIndex = errorsInserted;
		if (errorsInserted > (eventTableSize-1))	errorIndex = (eventTableSize-1);
		
		erasureIndex = erasuresInserted;
		if (erasuresInserted > (eventTableSize-1)) erasureIndex = (eventTableSize-1);

        errorsCorrected = rs255decode(y, u, z, paritySymbolCount);

        errorsFound = rs255compare(x,z,255);

        if (debugFlag==1) {
            polyPrint(e,255);
            debugFlag=0;
        }

/*
        if ((errorsCorrected != -1) &&((errorsInserted+erasuresInserted) != errorsCorrected)) {
            printf("\nNumber of errors do not match:  %3ld %3ld %3ld %3ld", 
                   errorsInserted, erasuresInserted, errorsCorrected, errorsFound);
            printf("\nEnter to continue ");
            scanf_c( "%c", &c );
        }
*/
        
/*There are two types of error/erasure patterns:
	(A1)	The error/erasure pattern has a correctable number of non-zero elements
	(A2)	The error/erasure pattern has weight higher than the theoretical limit
  There are two types of decoder output:
    (B1)    A certain number of errors/erasures were detected and decoded
	(B2)	The decoder detected a non-correctable error pattern
  Status of comparing codeword/errorword and decoder output:
    (C1)	All errors/erasures corrected (only related to B1)
	(C2)	Errors/erasures remain (only related to B1)
	(C3)    The decoder detected a non-correctable pattern (only related to B2)

  ABC
  111  A normal successful decoding
  112  An unacceptable malfunction in the decoder
  121  Not applicable (not possible)

  122  An unacceptable malfunction in the decoder
  211  May be possible, but not very often (if ever)
  212  A so-called decoding error, which happens
  221  A normal un-successful decoding
  222  Not applicable

*/
		if (errorsFound!=0) {
            eventCount[errorIndex][erasureIndex][1]++;
//            printf("*");
            if ((2*errorsInserted+erasuresInserted) <= paritySymbolCount) {
				correctableEventsNotCorrected++;
            }
			else 
				if (errorsCorrected == -1)
					notCorrectableEventsNotCorrected++;
				else
					notCorrectableEventsErrCorrected++;

        }
        else {
            eventCount[errorIndex][erasureIndex][0]++;
//            printf(".");
            if ((2*errorsInserted+erasuresInserted) <= paritySymbolCount) correctableEventsCorrected++;
			else notCorrectableEventsCorrected++;

        }
        i++;
		
		if ((i & 0xfff) == 0) {
            printf("\nCorrectable events with all errors corrected     %6ld", correctableEventsCorrected);
            printf("\nNon correctable events with all errors corrected %6ld", notCorrectableEventsCorrected);
            printf("\nCorrectable events with error in decoding        %6ld", correctableEventsNotCorrected);
            printf("\nNon correctable events unable to decode          %6ld", notCorrectableEventsNotCorrected);
            printf("\nNon correctable events with error in decoding    %6ld", notCorrectableEventsErrCorrected);
			printf("\n");
        }
        if ((i & 0xfff) == 0) printEventTable(txtFile, eventCount);
    } while ((nextWord(e,255)!=-1) && (i<10000000));

	if (0 == 0) {
        printf("\nCorrectable events with all errors corrected     %6ld", correctableEventsCorrected);
        printf("\nNon correctable events with all errors corrected %6ld", notCorrectableEventsCorrected);
        printf("\nCorrectable events with error in decoding        %6ld", correctableEventsNotCorrected);
        printf("\nNon correctable events unable to decode          %6ld", notCorrectableEventsNotCorrected);
        printf("\nNon correctable events with error in decoding    %6ld", notCorrectableEventsErrCorrected);
		printf("\n");
    }
    printEventTable(txtFile, eventCount);
  
	fclose(txtFile);   

    printf("\nEnter to continue ");
    scanf_s( "%c", &c );

    printf("\n");
    printf("\nSystematic error test terminated.");
    printf("\n");
}    


/* Reed-Solomon test bench main program. */
void main()
{    

	randomErrorErasureTest();
//	systematicErrorErasureTest();

}



